/*
 * Copyright 2010 John Schember <john@nachtimwald.com>
 *
 * This library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <QDir>
#include <QDomDocument>
#include <QDomElement>
#include <QFile>
#include <QFileInfo>

#include "epubreader.h"
#include "fileutils.h"
#include "zlibdecompressor.h"

EpubReader::EpubReader() : FormatReader()
{
    m_encoding = QTextCodec::codecForName("UTF-8");
}

EpubReader::~EpubReader()
{
    close();
}

bool EpubReader::open(QString const &filename)
{
    close();
    m_file = new QFile(filename);
    m_file->open(QFile::ReadOnly);
    
    readFileEntries();
    if (!validate()) {
        close();
        return false;
    }
    if (drmed()) {
        close();
        return false;
    }
    if (!readOpf()) {
        close();
        return false;
    }
    readNcx();

    return true;
}

void EpubReader::close()
{
    if (m_file) {
        if (m_file->isOpen()) {
            m_file->close();
        }
        delete m_file;
        m_file = 0;
    }

    m_manifest.clear();
    m_spine.clear();
    m_toc.clear();
    m_metadata.clear();
    m_entries.clear();
    m_emanifest.clear();
    m_opfDir = "";
    m_ncx = "";
}

bool EpubReader::extractBook(const QString &location)
{
    // The m_emanifest stores items relative to the OPF file.
    // The fullItemName is the full name used in the archive.
    QString fullItemName;
    QDir itemDir;

    m_manifest.clear();
    m_spine.clear();
    m_toc.clear();

    if (!FileUtils::removeDir(location)) {
        m_errorString = tr("Could not remove directory %1").arg(location);
        return false;
    }

    Q_FOREACH(QString item, m_emanifest.values()) {
        // The spine is relative to the OPF so we prepend the base location of the OPF
        // to the spine item. This is so it can be located in the archive.
        fullItemName = m_opfDir + "/" + item;

        itemDir = QFileInfo(location + "/" + item).absoluteDir();
        if (!itemDir.exists()) {
            if (!itemDir.mkpath(itemDir.absolutePath())) {
                m_errorString = tr("Could not create directory %1").arg(itemDir.absolutePath());
                return false;
            }
        }

        if (m_espine.contains(item)) {
            // Spine items need to be run though tidy before being saved.
            // When building the m_emanifest we already checked that the file is in
            // the archive so we won't check again before trying to extract it.
            QString text = m_encoding->toUnicode(extractFileEntry(m_entries.value(fullItemName)));
            text = generateXhtml(text);
            if (!writeXhtml(text, location, item)) {
                m_errorString = tr("Failed to write xhtml for %1").arg(item);
                return false;
            }
            m_spine.append(item);
        }
        else {
            QFile out(location + "/" + item);
            if (!out.open(QFile::WriteOnly | QFile::Truncate)) {
                m_errorString = tr("Unable to write file to %1").arg(location);
                return false;
            }
            out.write(extractFileEntry(m_entries.value(fullItemName)));
            out.close();
        }

        m_manifest.append(item);
    }

    m_toc = m_etoc;

    return true;
}

bool EpubReader::extractBookRaw(const QString &location)
{
    QString entryName;
    QDir entryDir;

    if (!FileUtils::removeDir(location)) {
        m_errorString = tr("Could not remove directory %1").arg(location);
        return false;
    }

    Q_FOREACH (EpubFileEntry entry, m_entries) {
        entryName = location + "/" + entry.filename();
        entryDir = QFileInfo(entryName).absoluteDir();

        if (!entryDir.exists()) {
            if (!entryDir.mkpath(entryDir.absolutePath())) {
                m_errorString = tr("Could not create directory %1").arg(entryDir.absolutePath());
                return false;
            }
        }

        QFile out(entryName);
        if (!out.open(QFile::WriteOnly | QFile::Truncate)) {
            m_errorString = tr("Unable to write file to %1").arg(location);
            return false;
        }
        out.write(extractFileEntry(entry));
        out.close();
    }

    return true;
}

bool EpubReader::validate()
{
    EpubFileEntry entry = m_entries.value("mimetype", EpubFileEntry());
    if (!entry.filename().isEmpty() && entry.uncompressedSize() != 0) {
        QString data = extractFileEntry(entry);
        if (data.startsWith("application/epub+zip")) {
            return true;
        }
    }
    m_errorString = tr("Not a valid EPUB file. Mimetype validation failed.");
    return false;
}

bool EpubReader::drmed()
{
    EpubFileEntry entry = m_entries.value("META-INF/encryption.xml", EpubFileEntry());
    if (entry.filename().isEmpty() && entry.uncompressedSize() == 0) {
        return true;
    }
    m_errorString = tr("File is encrypted.");
    return false;
}

bool EpubReader::readFileEntries()
{
    // Zip file entry signature.
    quint32 headerSignature = 0x4034b50;

    quint16 buf16 = 0;
    quint32 buf32 = 0;
    // The version required to decompress.
    quint16 version = 0;
    // General purpose flags.
    quint16 gpFlags = 0;
    quint16 filenameLength = 0;
    quint16 extraFieldLength = 0;
    quint64 seek = 0;

    QDataStream stream(m_file);
    stream.setByteOrder(QDataStream::LittleEndian);

    m_file->seek(0);

    // Read though the file creating a list of entries.
    // - 30 becuase that is the minimum size of a file entry
    // header.
    while (m_file->pos() < m_file->size() - 30) {
        EpubFileEntry entry;

        // Read the header signature.
        stream >> buf32;
        // We have a file entry
        if (buf32 == headerSignature) {
            // Minimum version needed to extract.
            stream >> buf16;
            version = buf16;

            // General purpose bit flag.
            stream >> gpFlags;

            // Compression method
            stream >> buf16;
            entry.setCompressionMethod(buf16);

            // Skip the last mod file time and last mod file data.
            m_file->seek(4 + m_file->pos());

            // Read the crc32.
            stream >> buf32;
            entry.setCrc32(buf32);

            // Read the compressed size of the data.
            stream >> buf32;
            entry.setCompressedSize(buf32);

            // Read the uncompressed size of the data.
            stream >> buf32;
            entry.setUncompressedSize(buf32);

            // Read the filename length.
            stream >> buf16;
            filenameLength = buf16;

            // Read the extra data field length.
            stream >> buf16;
            extraFieldLength = buf16;

            // Set the filename
            entry.setFilename(m_file->read(filenameLength));

            // Skip the extra field.
            m_file->seek(extraFieldLength + m_file->pos());

            // Set the data offset for the entry and move past it.
            entry.setDataOffset(m_file->pos());
            m_file->seek(entry.compressedSize() + m_file->pos());

            seek = m_file->pos();
            // Check if a data descriptor is avaliable for this entry.
            if ((gpFlags & (1 << 2)) != 0) {
                // crc32, compressed sise and uncompressed size
                // need to be read from the data descriptor when present.
                stream >> buf32;
                entry.setCrc32(buf32);
                stream >> buf32;
                entry.setCompressedSize(buf32);
                stream >> buf32;
                entry.setUncompressedSize(buf32);
                // Set to seek past the data descriptor;
                seek += 12;
            }

            // Move to the position of the next entry.
            if (seek < static_cast<quint64>(m_file->size())) {
                m_file->seek(seek);
            }
            else {
                m_file->seek(m_file->size() - 1);
            }

            // Various checks to see if we can use the entry.
            if (version != 10 && version != 20) {
                // Epub spec only allows for 10, 20 and 45.
                // TODO: Zip64 (45) support.
                continue;
            }
            // Epub spec only allows for uncompressed and Flate compression.
            if (entry.compressionMethod() != 0 && entry.compressionMethod() != 8) {
                continue;
            }
            if (entry.filename().isEmpty()) {
                continue;
            }
            // Typically means it's a directory indicator.
            if (entry.compressedSize() == 0 || entry.uncompressedSize() == 0) {
                continue;
            }

            m_entries.insert(entry.filename(), entry);
        }
        else {
            // File entries are always first in a zip file.
            // Quit searching for entires once they've all been read.
            break;
        }
    }

    return true;
}

bool EpubReader::readOpf()
{
    EpubFileEntry entry;
    QDomDocument doc;
    QString opfFile;
    QString bufs;
    QDomNodeList bufns;

    // Locate the OPF file.
    // Location of the OPF file is specified in the container.xml.
    entry = m_entries.value("META-INF/container.xml", EpubFileEntry());
    if (entry.filename().isEmpty()) {
        m_errorString = tr("Failed to parse OPF. container.xml missing.");
        return false;
    }
    doc.setContent(extractFileEntry(entry));
    if (doc.toString().isEmpty()) {
        m_errorString = tr("Failed to parse OPF. container.xml empty.");
        return false;
    }
    // The opf file should always be identified at <container><rootfiles><rootfile full-path="X">
    QDomNodeList roots = doc.firstChildElement("container").firstChildElement("rootfiles").elementsByTagName("rootfile");
    if (roots.count() >= 1) {
        QDomElement rootfile = roots.at(0).toElement();
        if (rootfile.attribute("media-type").contains("application/oebps-package+xml")) {
            opfFile = rootfile.attribute("full-path");
        }
        else {
            m_errorString = tr("Failed to parse OPF. OPF media type specified in container.xml is unsuppored.");
            return false;
        }
    }
    else {
        m_errorString = tr("Failed to parse OPF. container.xml does not specify a primary OPF file.");
        return false;
    }

    // Parse the OPF file.
    entry = m_entries.value(opfFile, EpubFileEntry());
    if (entry.filename() != opfFile) {
        m_errorString = tr("Failed to parse OPF. File is missing from archive.");
        return false;
    }
    QString opf = extractFileEntry(entry);
    if (opf.isEmpty()) {
        m_errorString = tr("Failed to parse OPF. OPF is empty.");
        return false;
    }

    doc.clear();
    doc.setContent(opf);
    QFileInfo opfInfo(opfFile);
    m_opfDir = opfInfo.path();

    // Read the manifest and spine from the opf file.
    // Create a dictionary of ids and their hrefs from the manifest.
    bufns = doc.firstChildElement("package").firstChildElement("manifest").elementsByTagName("item");
    for (uint i = 0; i < bufns.length(); i++) {
        QDomElement e = bufns.at(i).toElement();
        // Do not add the ncx to the manifest.
        if (e.attribute("href").endsWith(".ncx")) {
            m_ncx = e.attribute("href");
        }
        else {
            // Check that the file is actually in the archive before adding it to the manifest.
            if (m_entries.contains(m_opfDir + "/" + e.attribute("href"))) {
                m_emanifest.insert(e.attribute("id"), e.attribute("href"));
            }
        }
    }
    // Add the spine elements to the spine list.
    bufns = doc.firstChildElement("package").firstChildElement("spine").elementsByTagName("itemref");
    for (uint i = 0; i < bufns.length(); i++) {
        QDomElement e = bufns.at(i).toElement();
        bufs = m_emanifest.value(e.attribute("idref"), QString());
        // Only add spine items that we've included in the manifest (not the ncx for example).
        if (!bufs.isEmpty()) {
            m_espine.append(bufs);
        }
    }

    // Read the metadata.
    bufns = doc.firstChildElement("package").firstChildElement("metadata").childNodes();
    for (uint i = 0; i < bufns.length(); i++) {
        QDomElement e = bufns.at(i).toElement();
        m_metadata.value(e.nodeName().toLower().replace("dc:", ""), e.text());
    }

    return true;
}

void EpubReader::readNcx()
{
    // If the ncx is missing that's okay.
    if (m_ncx.isEmpty()) {
        return;
    }

    QString ncx = extractFileEntry(m_entries.value(m_opfDir + "/" + m_ncx));
    // Empty ncx is also okay.
    if (ncx.isEmpty()) {
        return;
    }

    QDomDocument doc;
    doc.setContent(ncx);

    m_etoc = readNcxNavPoints(doc.firstChildElement("ncx").firstChildElement("navMap").firstChildElement("navPoint"));
}

QList<TocItem> EpubReader::readNcxNavPoints(const QDomElement &element)
{
    QList<TocItem> items;
    TocItem toc;
    QDomElement e;

    QString name = element.firstChildElement("navLabel").firstChildElement("text").text();
    QString location = element.firstChildElement("content").attribute("src");

    if (!name.isEmpty() && !location.isEmpty()) {
        toc.setName(name);
        toc.setLocation(location);
        items.append(toc);
    }

    e = element.firstChildElement("navPoint");
    if (!e.isNull()) {
        items += readNcxNavPoints(e);
    }

    e = element.nextSiblingElement("navPoint");
    if (!e.isNull()) {
        items += readNcxNavPoints(e);
    }

    return items;
}

QByteArray EpubReader::extractFileEntry(EpubFileEntry entry)
{
    m_file->seek(entry.dataOffset());
    QByteArray data = m_file->read(entry.compressedSize());
    QByteArray out;

    // Epub spec requires the compression method to be set to either 0 or 8.
    // Also, any other compression method should be treated as being in error.
    if (entry.compressionMethod() == 0) {
        // No compression.
        out = data;
    }
    else if (entry.compressionMethod() == 8) {
        // Flate compression.
        out = ZlibDecompressor::decompressZip(data, entry.uncompressedSize());
    }
    else {
        return QByteArray();
    }
    if (out.length() != static_cast<int>(entry.uncompressedSize())) {
        return QByteArray();
    }
    
    return out;
}
