/*
 * Copyright 2010 John Schember <john@nachtimwald.com>
 *
 * This library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <QDir>
#include <QFileInfo>
#include <QRegExp>

#include "formatreader.h"

#include "buffio.h"

FormatReader::FormatReader()
{
    m_file = 0;
    m_encoding = 0;
    m_tidyDoc = 0;

    setupTidy();
}

FormatReader::~FormatReader()
{
    if (m_file) {
        m_file->close();
        delete m_file;
        m_file = 0;
    }

    tidyRelease(m_tidyDoc);
}

QString FormatReader::encoding()
{
    return m_encoding->name();
}

QStringList FormatReader::manifest()
{
    return m_manifest;
}

QStringList FormatReader::spine()
{
    return m_spine;
}

QList<TocItem> FormatReader::toc()
{
    return m_toc;
}

QHash<QString, QString> FormatReader::metadata()
{
    return m_metadata;
}

QString FormatReader::errorString()
{
    return m_errorString;
}

bool FormatReader::setEncoding(QString codec)
{
    m_encoding = QTextCodec::codecForName(codec.toLatin1());
    if (!m_encoding) {
        m_errorString = tr("Encoding %1 was not found").arg(codec);
        return false;
    }
    return true;
}

QString FormatReader::generateXhtml(const QString &html)
{
    TidyBuffer outBuffer;
    TidyBuffer errorBuffer;
    tidyBufInit(&outBuffer);
    tidyBufInit(&errorBuffer);
    QString xhtml = html;

    QRegExp xmlRegex = QRegExp("<\\?xml.*\\?>", Qt::CaseInsensitive, QRegExp::RegExp2);
    xmlRegex.setMinimal(true);
    xhtml = xhtml.replace(xmlRegex, "");

    tidySetErrorBuffer(m_tidyDoc, &errorBuffer);
    tidyParseString(m_tidyDoc, xhtml.toUtf8().constData());
    tidyCleanAndRepair(m_tidyDoc);
    tidySaveBuffer(m_tidyDoc, &outBuffer);

    xhtml = QString::fromUtf8((const char *)outBuffer.bp);

    tidyBufFree(&errorBuffer);
    tidyBufFree(&outBuffer);

    return xhtml;
}

bool FormatReader::writeXhtml(const QString &xhtml, const QString &baseLocation, const QString &filename)
{
    QFile file(baseLocation + "/" + filename);

    QFileInfo info(file.fileName());
    if (!info.absoluteDir().exists()) {
        QDir().mkpath(info.absolutePath());
    }

    if (file.open(QFile::WriteOnly | QFile::Truncate)) {
        file.write(xhtml.toUtf8());
        file.close();
        return true;
    }
    return false;
}

void FormatReader::setupTidy()
{
    m_tidyDoc = tidyCreate();

    tidySetCharEncoding(m_tidyDoc, "utf8");
    tidyOptSetValue(m_tidyDoc, TidyDoctype, "strict");
    tidyOptSetBool(m_tidyDoc, TidyDropEmptyParas, yes);
    tidyOptSetBool(m_tidyDoc, TidyFixBackslash, yes);
    tidyOptSetBool(m_tidyDoc, TidyFixComments, yes);
    tidyOptSetBool(m_tidyDoc, TidyIndentAttributes, yes);
    tidyOptSetInt(m_tidyDoc, TidyIndentContent, TidyAutoState);
    tidyOptSetInt(m_tidyDoc, TidyIndentSpaces, 4);
    tidyOptSetBool(m_tidyDoc, TidyMark, no);
    tidyOptSetInt(m_tidyDoc, TidyWrapLen, 80);
    tidyOptSetBool(m_tidyDoc, TidyXhtmlOut, yes);
    tidyOptSetBool(m_tidyDoc, TidyXmlOut, no);
}
