/*
 * Copyright 2010 John Schember <john@nachtimwald.com>
 *
 * This library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <QByteArray>
#include <QDataStream>
#include <QDir>
#include <QFile>

#include "fileutils.h"
#include "palmdoc.h"
#include "palmdocreader.h"
#include "textconverter.h"

const QString PalmDocReader::IDENTITY = "TEXtREAd";
const quint16 PalmDocReader::MAX_RECORD_SIZE = 4096;

PalmDocReader::PalmDocReader() : PdbReader()
{
    m_record0 = 0;
    m_encoding = QTextCodec::codecForName("Windows-1252");
}

PalmDocReader::~PalmDocReader()
{
    close();
}

bool PalmDocReader::open(QString const &filename)
{
    close();

    m_file = new QFile(filename);
    m_file->open(QFile::ReadOnly);
    readHeader();

    if (!checkIdentity(IDENTITY)) {
        close();
        return false;
    }

    readRecord0();

    if (!validate()) {
        close();
        return false;
    }

    readMetadata();

    return true;
}

void PalmDocReader::close()
{
    if (m_file) {
        if (m_file->isOpen()) {
            m_file->close();
        }
        delete m_file;
        m_file = 0;
    }

    if (m_record0) {
        delete m_record0;
        m_record0 = 0;
    }

    m_manifest.clear();
    m_spine.clear();
    m_toc.clear();
    m_metadata.clear();
}

bool PalmDocReader::extractBook(const QString &location)
{
    QString text;

    m_manifest.clear();

    if (!FileUtils::removeDir(location)) {
        m_errorString = tr("Could not remove directory %1").arg(location);
        return false;
    }
    QDir locationDir(location);
    if (!locationDir.exists(locationDir.absolutePath())) {
        if (!locationDir.mkpath(locationDir.absolutePath())) {
            m_errorString = tr("Could not create directory %1").arg(locationDir.absolutePath());
            return false;
        }
    }

    for (int i = 1; i < m_record0->recordCount() + 1; i++) {
        text += getText(i);
    }

    text = TextConverter::toHtml(text);
    text = generateXhtml(text);

    if (!writeXhtml(text, location, "index.html")) {
        m_errorString = tr("Could not write file index.html to %1").arg(location);
        return false;
    }

    m_spine.append("index.html");
    m_manifest.append("index.html");
    return true;
}

bool PalmDocReader::extractBookRaw(const QString &location)
{
    QString text;

    if (!FileUtils::removeDir(location)) {
        m_errorString = tr("Could not remove directory %1").arg(location);
        return false;
    }
    QDir locationDir(location);
    if (!locationDir.exists(locationDir.absolutePath())) {
        if (!locationDir.mkpath(locationDir.absolutePath())) {
            m_errorString = tr("Could not create directory %1").arg(locationDir.absolutePath());
            return false;
        }
    }

    for (int i = 1; i < m_record0->recordCount() + 1; i++) {
        text += getText(i);
    }

    QFile out(location + "/index.txt");
    if (!out.open(QFile::WriteOnly | QFile::Truncate)) {
        m_errorString = tr("Unable to write file to %1").arg(location);
        return false;
    }
    out.write(text.toUtf8());
    out.close();

    return true;
}

void PalmDocReader::readRecord0()
{
    m_record0 = new PalmDocHeader();
    quint16 buf16;
    quint32 buf32;

    QByteArray rawData = getSectionData(0);
    QDataStream stream(rawData);
    stream.setByteOrder(QDataStream::BigEndian);

    stream >> buf16;
    m_record0->setCompressed(buf16 == 2);
    // Skip the unused section that should always be 0.
    stream >> buf16;
    stream >> buf32;
    m_record0->setUncompressedLength(buf32);
    stream >> buf16;
    m_record0->setRecordCount(buf16);
    // Max record size will always be 4096. We are going to store and check it later
    // to be sure the file isn not damaged.
    stream >> buf16;
    m_record0->setMaxRecordSize(buf16);
}

bool PalmDocReader::validate()
{
    // Check that there is text.
    if (m_record0->recordCount() < 1) {
        m_errorString = tr("Book contains no text.");
        return false;
    }

    // Ensure the file is not damanged by checking the stored max record size.
    if (m_record0->maxRecordSize() != MAX_RECORD_SIZE) {
        m_errorString = tr("Invalid max record size of %1").arg(m_record0->maxRecordSize());
        return false;
    }

    return true;
}

void PalmDocReader::readMetadata()
{
    m_metadata.insert("title", m_pdbHeader->title());
}

QString PalmDocReader::getText(quint16 section)
{
    QByteArray data = getSectionData(section);

    if (m_record0->Compressed()) {
        data = PalmDoc::decompress(data);
    }

    return m_encoding->toUnicode(data);
}
