/*
 * Copyright 2010 John Schember <john@nachtimwald.com>
 *
 * This library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <QDataStream>

#include "pdbreader.h"

PdbReader::PdbReader() : FormatReader()
{
}

PdbReader::~PdbReader()
{
    if (m_pdbHeader) {
        delete m_pdbHeader;
        m_pdbHeader = 0;
    }
}

QString PdbReader::getIdenity(const QString &filename)
{
    QFile file(filename);
    file.open(QFile::ReadOnly);
    QString ident = getIdenity(&file);
    file.close();

    return ident;
}

QString PdbReader::getIdenity(QFile *file)
{
    file->seek(60);
    return QString(file->read(8));
}

bool PdbReader::checkIdentity(const QString &identity)
{
    if (m_pdbHeader->identity() != identity) {
        m_errorString = tr("Detected an identity of %1 but an identity of %2 is requried.").arg(identity).arg(m_pdbHeader->identity());
        return false;
    }
    return true;
}

QByteArray PdbReader::getSectionData(quint16 number)
{
    if (number >= m_pdbHeader->sectionCount()) {
        m_errorString = tr("Index %1 is out of range").arg(number);
        return QByteArray();
    }

    m_file->seek(m_pdbHeader->sectionOffsets().at(number));
    return m_file->read(m_pdbHeader->sectionLengths().at(number));
}

void PdbReader::readHeader()
{
    m_pdbHeader = new PdbHeader();
    QDataStream stream(m_file);
    stream.setByteOrder(QDataStream::BigEndian);

    // Title of the archive.
    m_file->seek(0);
    m_pdbHeader->setTitle(QString(m_file->read(32)).replace("\0", ""));

    // Format of the data contained within the archive.
    m_pdbHeader->setIdentity(getIdenity(m_file));

    // Total number of sections the data is divided into.
    m_file->seek(76);
    quint16 sectionCount = 0;
    stream >> sectionCount;
    m_pdbHeader->setSectionCount(sectionCount);

    // Offset to the start of each section within the file.
    QList<quint32> sectionOffsets;
    quint32 offset = 0;
    for (int i = 0; i < sectionCount; i++) {
        m_file->seek(78 + (i * 8));
        stream >> offset;
        sectionOffsets.append(offset);
    }
    m_pdbHeader->setSectionOffsets(sectionOffsets);

    // Setup the section lengths.
    QList<quint16> sectionLengths;
    for (int i = 0; i < sectionCount; i++) {
        if (i + 1 == sectionCount) {
            sectionLengths.append(m_file->size() - sectionOffsets.at(i));
        }
        else {
            sectionLengths.append(sectionOffsets.at(i + 1) - sectionOffsets.at(i));
        }
    }
    m_pdbHeader->setSectionLengths(sectionLengths);
}
