/*
 * Copyright 2010 John Schember <john@nachtimwald.com>
 *
 * This library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <QScopedPointer>

#include <zconf.h>

#include "zlibdecompressor.h"

const int ZlibDecompressor::MAX_BUFFER_SIZE = 300;

ZlibDecompressor::ZlibDecompressor()
{
    m_zStream.zalloc = Z_NULL;
    m_zStream.zfree = Z_NULL;
    m_zStream.opaque = Z_NULL;
    m_zStream.avail_in = 0;
    m_zStream.next_in = Z_NULL;

    inflateInit(&m_zStream);
}

ZlibDecompressor::~ZlibDecompressor()
{
    inflateEnd(&m_zStream);
}

QByteArray ZlibDecompressor::decompress(const QByteArray &data)
{
    QScopedPointer<char> buf(new char[MAX_BUFFER_SIZE]);
    QByteArray out;
    int ret;

    m_zStream.avail_in = data.length();
    if (m_zStream.avail_in == 0) {
        return QByteArray();
    }
    m_zStream.next_in = reinterpret_cast<Bytef *>(const_cast<char *>(data.data()));

    do {
        m_zStream.avail_out = MAX_BUFFER_SIZE;
        m_zStream.next_out = reinterpret_cast<Bytef *>(buf.data());
        ret = inflate(&m_zStream, Z_FULL_FLUSH);
        if (ret == Z_NEED_DICT || ret == Z_DATA_ERROR || ret == Z_MEM_ERROR) {
            return QByteArray();
        }
        out.append(buf.data(), MAX_BUFFER_SIZE - m_zStream.avail_out);

    } while (ret != Z_STREAM_END && m_zStream.avail_out == 0);

    return out;
}

QByteArray ZlibDecompressor::decompressZip(const QByteArray &data, quint64 uncompessedSize)
{
    z_stream zStream;
    QScopedPointer<char> buf(new char[uncompessedSize]);
    int ret;
    QByteArray out;

    zStream.zalloc = Z_NULL;
    zStream.zfree = Z_NULL;
    zStream.opaque = Z_NULL;
    zStream.avail_in = 0;
    zStream.next_in = Z_NULL;

    inflateInit2(&zStream, -MAX_WBITS);

    zStream.avail_in = data.length();
    zStream.next_in = reinterpret_cast<Bytef *>(const_cast<char *>(data.data()));

    zStream.avail_out = uncompessedSize;
    zStream.next_out = reinterpret_cast<Bytef *>(buf.data());

    // inflate can do everything it needs in one call because we
    // are using a buffer the size of the uncompessed data.
    ret = inflate(&zStream, Z_SYNC_FLUSH);
    out.append(buf.data(), uncompessedSize - zStream.avail_out);

    inflateEnd(&zStream);

    if (ret != Z_STREAM_END) {
        return QByteArray();
    }

    return out;
}
