/*
 * Copyright 2010 John Schember <john@nachtimwald.com>
 *
 * This library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <QByteArray>
#include <QDataStream>

#include <zlib.h>

#include "fileutils.h"
#include "textconverter.h"
#include "ztxtreader.h"

const QString ZTxtReader::IDENTITY = "zTXTGPlm";
const quint8 ZTxtReader::SUPPORTED_MAJOR_VERSION = 1;
const quint8 ZTxtReader::SUPPORTED_MINOR_VERSION = 44;

ZTxtReader::ZTxtReader() : PdbReader()
{
    m_record0 = 0;
    m_encoding = QTextCodec::codecForName("Windows-1252");
    m_cMode1 = true;
}

ZTxtReader::~ZTxtReader()
{
    close();
}

bool ZTxtReader::open(QString const &filename)
{
    close();

    m_file = new QFile(filename);
    m_file->open(QFile::ReadOnly);
    readHeader();

    if (!checkIdentity(IDENTITY)) {
        close();
        return false;
    }

    readRecord0();

    if (!validate()) {
        close();
        return false;
    }

    // Set compression mode.
    m_cMode1 = ((m_record0->flags() & 0x01) == 0);

    readMetadata();

    return true;
}

void ZTxtReader::close()
{
    if (m_file) {
        if (m_file->isOpen()) {
            m_file->close();
        }
        delete m_file;
        m_file = 0;
    }

    if (m_record0) {
        delete m_record0;
        m_record0 = 0;
    }

    m_manifest.clear();
    m_spine.clear();
    m_toc.clear();
    m_metadata.clear();
}

bool ZTxtReader::extractBook(const QString &location)
{
    m_manifest.clear();

    if (!FileUtils::removeDir(location)) {
        m_errorString = tr("Could not remove directory %1").arg(location);
        return false;
    }
    QDir locationDir(location);
    if (!locationDir.exists(locationDir.absolutePath())) {
        if (!locationDir.mkpath(locationDir.absolutePath())) {
            m_errorString = tr("Could not create directory %1").arg(locationDir.absolutePath());
            return false;
        }
    }

    QString text = getText();

    text = TextConverter::toHtml(text);
    text = generateXhtml(text);

    if (!writeXhtml(text, location, "index.html")) {
        m_errorString = tr("Could not write file index.html to %1").arg(location);
        return false;
    }

    m_spine.append("index.html");
    m_manifest.append("index.html");
    return true;
}

bool ZTxtReader::extractBookRaw(const QString &location)
{
    if (!FileUtils::removeDir(location)) {
        m_errorString = tr("Could not remove directory %1").arg(location);
        return false;
    }
    QDir locationDir(location);
    if (!locationDir.exists(locationDir.absolutePath())) {
        if (!locationDir.mkpath(locationDir.absolutePath())) {
            m_errorString = tr("Could not create directory %1").arg(locationDir.absolutePath());
            return false;
        }
    }

    QString text = getText();

    QFile out(location + "/index.txt");
    if (!out.open(QFile::WriteOnly | QFile::Truncate)) {
        m_errorString = tr("Unable to write file to %1").arg(location);
        return false;
    }
    out.write(text.toUtf8());
    out.close();

    return true;
}

void ZTxtReader::readRecord0()
{
    m_record0 = new ZTxtHeader();
    quint8 buf8;
    quint16 buf16;
    quint32 buf32;

    QByteArray rawData = getSectionData(0);
    QDataStream stream(rawData);
    stream.setByteOrder(QDataStream::BigEndian);

    stream >> buf16;
    m_record0->setVersion(buf16);
    stream >> buf16;
    m_record0->setNumRecords(buf16);
    stream >> buf32;
    m_record0->setSize(buf32);
    stream >> buf16;
    m_record0->setRecordSize(buf16);
    stream >> buf16;
    m_record0->setNumBookmarks(buf16);
    stream >> buf16;
    m_record0->setBookmarkRecord(buf16);
    stream >> buf16;
    m_record0->setNumAnnotation(buf16);
    stream >> buf16;
    m_record0->setAnnotationRecord(buf16);
    stream >> buf8;
    m_record0->setFlags(buf8);
    // Reserved data
    stream >> buf8;
    stream >> buf32;
    m_record0->setCrc32(buf32);
}

bool ZTxtReader::validate()
{
    // Check that there is text.
    if (m_record0->numRecords() < 1) {
        m_errorString = tr("Book contains no text.");
        return false;
    }

    // Check version.
    quint8 vmajor = (m_record0->version() & 0x0000FF00) >> 8;
    quint8 vminor = m_record0->version() & 0x000000FF;
    if (vmajor != SUPPORTED_MAJOR_VERSION && vminor < SUPPORTED_MINOR_VERSION) {
        m_errorString = tr("Unsupported ztxt version (%1.%2). Only versions newer than %3.%4 are supported.").arg(vmajor).arg(vminor).arg(SUPPORTED_MAJOR_VERSION).arg(SUPPORTED_MINOR_VERSION);
        return false;
    }

    // Check crc32.
    ulong crc = crc32(0L, Z_NULL, 0);
    for (quint16 i = 1; i <= m_record0->numRecords(); i++) {
        QByteArray data = getSectionData(i);
        crc = crc32(crc, reinterpret_cast<Bytef *>(const_cast<char *>(data.data())), data.length());
    }
    if (m_record0->crc32() != 0 && crc != m_record0->crc32()) {
        m_errorString = tr("File is damanged. Crc32 validation failed. Got: %1. Expected: %2.").arg(crc).arg(m_record0->crc32());
        return false;
    }

    return true;
}

void ZTxtReader::readMetadata()
{
    m_metadata.insert("title", m_pdbHeader->title());
}

QString ZTxtReader::getText()
{
    ZlibDecompressor decompressor;
    QByteArray data;

    if (m_cMode1) {
        for (int i = 1; i <= m_record0->numRecords(); i++) {
            QByteArray sData;
            sData = getSectionData(i);
            data += decompressor.decompress(sData);
        }
    }
    else {
        for (int i = 1; i <= m_record0->numRecords(); i++) {
            data += getSectionData(i);
        }
        data = decompressor.decompress(data);
    }

    return m_encoding->toUnicode(data);
}
